//  icalBuddy.m
//
//  Created by Ali Rantakari on 17 June 2008

/*
The MIT License

Copyright (c) 2008 Ali Rantakari

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/


#import <Foundation/Foundation.h>
#import <CalendarStore/CalendarStore.h>
#import <AppKit/AppKit.h>





const int VERSION_MAJOR = 1;
const int VERSION_MINOR = 0;
const int VERSION_BUILD = 3;



#define kAppSiteURLPrefix @"http://hasseg.org/icalBuddy/"
#define kVersionCheckURL [NSURL URLWithString:@"http://hasseg.org/icalBuddy/?versioncheck=y"]
#define kDefaultTimeFormatString @"%H:%M"
#define kDefaultDateTimeFormatString @"%Y-%m-%d %H:%M:%S %z"



enum eventPrintContext
{
	EVENTS_TODAY = 1, 				// in the contex of today
	EVENTS_TODAY_NOCALENDAR = 2, 	// in the context of today, calendar-agnostic
	EVENTS_WHENEVER = 3 			// in the context of whenever
} EventPrintContext;

enum taskPrintContext
{
	TASKS = 1,						// default
	TASKS_NOCALENDAR = 2			// calendar-agnostic
} TaskPrintContext;






// helper method
NSString* versionNumber()
{
	return [NSString stringWithFormat:@"%d.%d.%d", VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD];
}


// helper method
// 		from: http://www.sveinbjorn.org/objectivec_stdout
// 		(modified to use non-deprecated version of writeToFile:...)
/*
reminder: string encoding consts:
enum {
   NSASCIIStringEncoding = 1,
   NSNEXTSTEPStringEncoding = 2,
   NSJapaneseEUCStringEncoding = 3,
   NSUTF8StringEncoding = 4,
   NSISOLatin1StringEncoding = 5,
   NSSymbolStringEncoding = 6,
   NSNonLossyASCIIStringEncoding = 7,
   NSShiftJISStringEncoding = 8,
   NSISOLatin2StringEncoding = 9,
   NSUnicodeStringEncoding = 10,
   NSWindowsCP1251StringEncoding = 11,
   NSWindowsCP1252StringEncoding = 12,
   NSWindowsCP1253StringEncoding = 13,
   NSWindowsCP1254StringEncoding = 14,
   NSWindowsCP1250StringEncoding = 15,
   NSISO2022JPStringEncoding = 21,
   NSMacOSRomanStringEncoding = 30,
   NSUTF16BigEndianStringEncoding = 0x90000100,
   NSUTF16LittleEndianStringEncoding = 0x94000100,
   NSUTF32StringEncoding = 0x8c000100,
   NSUTF32BigEndianStringEncoding = 0x98000100,
   NSUTF32LittleEndianStringEncoding = 0x9c000100,
   NSProprietaryStringEncoding = 65536
};
*/
void NSPrint (NSString *str)
{
     [str writeToFile: @"/dev/stdout" atomically: NO encoding: NSISOLatin2StringEncoding error: NULL];
}





// helper method: compare three-part version number strings (e.g. "1.0.3")
NSComparisonResult versionNumberCompare(NSString *first, NSString *second)
{
	if (first != nil && second != nil)
	{
		int i;
		
		NSMutableArray *firstComponents = [NSMutableArray arrayWithCapacity:3];
		[firstComponents addObjectsFromArray:[first componentsSeparatedByString:@"."]];
		
		NSMutableArray *secondComponents = [NSMutableArray arrayWithCapacity:3];
		[secondComponents addObjectsFromArray:[second componentsSeparatedByString:@"."]];
		
		if ([firstComponents count] != [secondComponents count])
		{
			NSMutableArray *shorter;
			NSMutableArray *longer;
			if ([firstComponents count] > [secondComponents count])
			{
				shorter = secondComponents;
				longer = firstComponents;
			}
			else
			{
				shorter = firstComponents;
				longer = secondComponents;
			}
			
			NSUInteger countDiff = [longer count] - [shorter count];
			
			for (i = 0; i < countDiff; i++)
				[shorter addObject:@"0"];
		}
		
		for (i = 0; i < [firstComponents count]; i++)
		{
			int firstComponentIntVal = [[firstComponents objectAtIndex:i] intValue];
			int secondComponentIntVal = [[secondComponents objectAtIndex:i] intValue];
			if (firstComponentIntVal < secondComponentIntVal)
				return NSOrderedAscending;
			else if (firstComponentIntVal > secondComponentIntVal)
				return NSOrderedDescending;
		}
		return NSOrderedSame;
	}
	else
		return NSOrderedSame;
}




// create an NSSet from a comma-separated string,
// trimming whitespace from around each string component
NSSet* setFromCommaSeparatedStringTrimmingWhitespace(NSString *str)
{
	if (str != nil)
	{
		NSMutableSet *set = [NSMutableSet setWithCapacity:10];
		NSArray *arr = [str componentsSeparatedByString:@","];
		NSString *component;
		for (component in arr)
			[set addObject:[component stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]]];
		return set;
	}
	return [NSSet set];
}


// create an NSArray from a comma-separated string,
// trimming whitespace from around each string component
NSArray* arrayFromCommaSeparatedStringTrimmingWhitespace(NSString *str)
{
	if (str != nil)
	{
		NSMutableArray *retArr = [NSMutableArray arrayWithCapacity:10];
		NSArray *arr = [str componentsSeparatedByString:@","];
		NSString *component;
		for (component in arr)
			[retArr addObject:[component stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]]];
		return retArr;
	}
	return [NSArray array];
}




// sort function for sorting tasks by priority:
// sort numerically except treat CalPriorityNone (0) as a special case
NSInteger prioritySort(id task1, id task2, void *context)
{
    if ([task1 priority] < [task2 priority])
	{
		if ([task1 priority] == CalPriorityNone)
			return NSOrderedDescending;
		else
			return NSOrderedAscending;
	}
    else if ([task1 priority] > [task2 priority])
		if ([task2 priority] == CalPriorityNone)
			return NSOrderedAscending;
		else
			return NSOrderedDescending;
    else
        return NSOrderedSame;
}


// whether the two specified dates represent the same calendar day
BOOL datesRepresentSameDay (NSCalendarDate *date1, NSCalendarDate *date2)
{
	return ([date1 yearOfCommonEra] == [date2 yearOfCommonEra] && [date1 monthOfYear] == [date2 monthOfYear] && [date1 dayOfMonth] == [date2 dayOfMonth]);
}



// helper method
BOOL shouldPrintProperty(NSString *propertyName, NSSet *inclusionsSet, NSSet *exclusionsSet)
{
	if (propertyName == nil || (inclusionsSet == nil && exclusionsSet == nil))
		return YES;
	
	BOOL retVal = YES;
	
	if (inclusionsSet != nil)
		if (![inclusionsSet containsObject:propertyName])
			retVal = NO;
	
	if (retVal == YES && exclusionsSet != nil)
		if ([exclusionsSet containsObject:propertyName] || [exclusionsSet containsObject:@"*"])
			retVal = NO;
	
	return retVal;
}





// pretty-prints out the specified event with the specified context
void printCalEvent (CalEvent *event, enum eventPrintContext context,
					NSString *timeFormatStr, NSString *dateTimeFormatStr,
					NSSet *includedProps, NSSet *excludedProps)
{
	if (event != nil)
	{
		NSCalendarDate *now = [NSCalendarDate calendarDate];
		
		// title
		if (context == EVENTS_TODAY_NOCALENDAR)
			NSPrint([NSString stringWithFormat: @"* %@\n", [event title]]);
		else
			NSPrint([NSString stringWithFormat: @"* %@ (%@)\n", [event title],[[event calendar] title]]);
		
		// location
		if (shouldPrintProperty(@"location", includedProps, excludedProps))
			if ([event location] != nil && ![[[event location] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:@""])
				NSPrint([NSString stringWithFormat: @"    location: %@\n", [event location]]);
		
		// notes
		if (shouldPrintProperty(@"notes", includedProps, excludedProps))
			if ([event notes] != nil && ![[[event notes] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:@""])
				NSPrint([NSString stringWithFormat: @"    notes: %@\n", [[event notes] stringByReplacingOccurrencesOfString:@"\n" withString:@"\n           "]]);
		
		// url
		if (shouldPrintProperty(@"url", includedProps, excludedProps))
			if ([event url] != nil)
				NSPrint([NSString stringWithFormat: @"    url: %@\n", [event url]]);
		
		// start/end dates & times
		if (shouldPrintProperty(@"startDate", includedProps, excludedProps) &&
			shouldPrintProperty(@"endDate", includedProps, excludedProps))
			if (![event isAllDay])
			{
				if ([[event startDate] isEqualToDate:[event endDate]])
				{
					if (context == EVENTS_TODAY || context == EVENTS_TODAY_NOCALENDAR)
						NSPrint([NSString stringWithFormat: @"    %@\n", [[event startDate] descriptionWithCalendarFormat:(timeFormatStr != nil)?timeFormatStr:kDefaultTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
					else
						NSPrint([NSString stringWithFormat: @"    %@\n", [[event startDate] descriptionWithCalendarFormat:(dateTimeFormatStr != nil)?dateTimeFormatStr:kDefaultDateTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
				}
				else
				{
					if (context == EVENTS_TODAY || context == EVENTS_TODAY_NOCALENDAR)
					{
						BOOL startsToday = datesRepresentSameDay(now, [[event startDate] dateWithCalendarFormat:nil timeZone:nil]);
						BOOL endsToday = datesRepresentSameDay(now, [[event endDate] dateWithCalendarFormat:nil timeZone:nil]);
						if (startsToday && endsToday)
							NSPrint([NSString stringWithFormat: @"    %@ - %@\n", [[event startDate] descriptionWithCalendarFormat:(timeFormatStr != nil)?timeFormatStr:kDefaultTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]], [[event endDate] descriptionWithCalendarFormat:(timeFormatStr != nil)?timeFormatStr:kDefaultTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
						else if (startsToday)
							NSPrint([NSString stringWithFormat: @"    %@ - ...\n", [[event startDate] descriptionWithCalendarFormat:(timeFormatStr != nil)?timeFormatStr:kDefaultTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
						else if (endsToday)
							NSPrint([NSString stringWithFormat: @"    ... - %@\n", [[event endDate] descriptionWithCalendarFormat:(timeFormatStr != nil)?timeFormatStr:kDefaultTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
					}
					else
						NSPrint([NSString stringWithFormat: @"    %@ - %@\n", [[event startDate] descriptionWithCalendarFormat:(dateTimeFormatStr != nil)?dateTimeFormatStr:kDefaultDateTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]], [[event endDate] descriptionWithCalendarFormat:(dateTimeFormatStr != nil)?dateTimeFormatStr:kDefaultDateTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
				}
			}
	}
}



// pretty-prints out the specified task with the specified context
void printCalTask (CalTask *task, enum eventPrintContext context,
					NSString *timeFormatStr, NSString *dateTimeFormatStr,
					NSSet *includedProps, NSSet *excludedProps)
{
	if (task != nil)
	{
		NSCalendarDate *now = [NSCalendarDate calendarDate];
		
		// title
		if (context == TASKS_NOCALENDAR)
			NSPrint([NSString stringWithFormat: @"* %@\n", [task title]]);
		else
			NSPrint([NSString stringWithFormat: @"* %@ (%@)\n", [task title],[[task calendar] title]]);
		
		// notes
		if (shouldPrintProperty(@"notes", includedProps, excludedProps))
			if ([task notes] != nil && ![[[task notes] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceCharacterSet]] isEqualToString:@""])
				NSPrint([NSString stringWithFormat: @"    notes: %@\n", [[task notes] stringByReplacingOccurrencesOfString:@"\n" withString:@"\n           "]]);
		
		// url
		if (shouldPrintProperty(@"url", includedProps, excludedProps))
			if ([task url] != nil)
				NSPrint([NSString stringWithFormat: @"    url: %@\n", [task url]]);
		
		// due date
		if (shouldPrintProperty(@"dueDate", includedProps, excludedProps))
			if ([task dueDate] != nil)
			{
				BOOL dueToday = datesRepresentSameDay(now, [[task dueDate] dateWithCalendarFormat:nil timeZone:nil]);
				if (dueToday)
					NSPrint([NSString stringWithFormat: @"    due: %@\n", [[task dueDate] descriptionWithCalendarFormat:(timeFormatStr != nil)?timeFormatStr:kDefaultTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
				else
					NSPrint([NSString stringWithFormat: @"    due: %@\n", [[task dueDate] descriptionWithCalendarFormat:(dateTimeFormatStr != nil)?dateTimeFormatStr:kDefaultDateTimeFormatString timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]] ]);
			}
		
		// priority
		if (shouldPrintProperty(@"priority", includedProps, excludedProps))
			if ([task priority] != CalPriorityNone)
			{
				if ([task priority] == CalPriorityHigh)
					NSPrint(@"    priority: high\n");
				else if ([task priority] == CalPriorityMedium)
					NSPrint(@"    priority: medium\n");
				else if ([task priority] == CalPriorityLow)
					NSPrint(@"    priority: low\n");
				else
					NSPrint([NSString stringWithFormat: @"    priority: %d\n", [task priority]]);
			}
	}
}










int main(int argc, char *argv[])
{
	NSAutoreleasePool *autoReleasePool = [[NSAutoreleasePool alloc] init];
	
	
	
	// variables for arguments
	NSString *arg_output = nil;
	BOOL arg_separateCals = NO;
	NSArray *arg_includeCals = nil;
	NSArray *arg_excludeCals = nil;
	BOOL arg_updatesCheck = NO;
	BOOL arg_printVersion = NO;
	NSString *arg_timeFormatStr = nil;
	NSString *arg_dateTimeFormatStr = nil;
	NSSet *arg_includedEventProperties = nil;
	NSSet *arg_excludedEventProperties = nil;
	NSSet *arg_includedTaskProperties = nil;
	NSSet *arg_excludedTaskProperties = nil;
	BOOL arg_includeOnlyEventsFromNowOn = NO;
	
	// get arguments
	if (argc > 1)
		arg_output = [NSString stringWithCString: argv[argc-1] encoding: NSASCIIStringEncoding];
	
	int i;
	for (i = 0; i < argc; i++)
	{
		if ((strcmp(argv[i], "-s") == 0) || (strcmp(argv[i], "--separate") == 0))
			arg_separateCals = YES;
		if ((strcmp(argv[i], "-u") == 0) || (strcmp(argv[i], "--checkForUpdates") == 0))
			arg_updatesCheck = YES;
		if ((strcmp(argv[i], "-V") == 0) || (strcmp(argv[i], "--version") == 0))
			arg_printVersion = YES;
		if ((strcmp(argv[i], "-n") == 0) || (strcmp(argv[i], "--includeOnlyEventsFromNowOn") == 0))
			arg_includeOnlyEventsFromNowOn = YES;
	}
	
	NSString *includeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"ic"];
	if (includeCalsStr == nil)
		includeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-includeCals"];
	if (includeCalsStr != nil)
		arg_includeCals = arrayFromCommaSeparatedStringTrimmingWhitespace(includeCalsStr);
	
	NSString *excludeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"ec"];
	if (excludeCalsStr == nil)
		excludeCalsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-excludeCals"];
	if (excludeCalsStr != nil)
		arg_excludeCals = arrayFromCommaSeparatedStringTrimmingWhitespace(excludeCalsStr);
	
	arg_timeFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"tf"];
	if (arg_timeFormatStr == nil)
		arg_timeFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-timeFormat"];
	
	arg_dateTimeFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"dtf"];
	if (arg_dateTimeFormatStr == nil)
		arg_dateTimeFormatStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-dateTimeFormat"];
	
	NSString *includeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"iep"];
	if (includeEventPropsStr == nil)
		includeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-includeEventProps"];
	if (includeEventPropsStr != nil)
		arg_includedEventProperties = setFromCommaSeparatedStringTrimmingWhitespace(includeEventPropsStr);
	
	NSString *excludeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"eep"];
	if (excludeEventPropsStr == nil)
		excludeEventPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-excludeEventProps"];
	if (excludeEventPropsStr != nil)
		arg_excludedEventProperties = setFromCommaSeparatedStringTrimmingWhitespace(excludeEventPropsStr);
	
	NSString *includeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"itp"];
	if (includeTaskPropsStr == nil)
		includeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-includeTaskProps"];
	if (includeTaskPropsStr != nil)
		arg_includedTaskProperties = setFromCommaSeparatedStringTrimmingWhitespace(includeTaskPropsStr);
	
	NSString *excludeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"etp"];
	if (excludeTaskPropsStr == nil)
		excludeTaskPropsStr = [[NSUserDefaults standardUserDefaults] stringForKey:@"-excludeTaskProps"];
	if (excludeTaskPropsStr != nil)
		arg_excludedTaskProperties = setFromCommaSeparatedStringTrimmingWhitespace(excludeTaskPropsStr);
	
	
	
	
	
	
	
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print version and exit
	// ------------------------------------------------------------------
	if (arg_printVersion)
	{
		NSPrint([NSString stringWithFormat:@"%@\n", versionNumber()]);
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// check for updates
	// ------------------------------------------------------------------
	else if (arg_updatesCheck)
	{
		NSPrint(@"Checking for updates... ");
		
		NSURL *url = kVersionCheckURL;
		NSURLRequest *request = [NSURLRequest requestWithURL:url cachePolicy:NSURLRequestReloadIgnoringCacheData timeoutInterval:10.0];
		
		NSHTTPURLResponse *response;
		NSError *error;
		[NSURLConnection sendSynchronousRequest:request returningResponse:&response error:&error];
		
		if (error == nil && response != nil)
		{
			NSInteger statusCode = [response statusCode];
			if (statusCode >= 400)
			{
				NSPrint([NSString stringWithFormat:@"...HTTP connection failed.\n\nStatus code %d: \"%@\"\n\n",
									statusCode, [NSHTTPURLResponse localizedStringForStatusCode:statusCode]]);
			}
			else
			{
				NSString *latestVersionString = [[response allHeaderFields] valueForKey:@"Orghassegsoftwarelatestversion"];
				NSString *currentVersionString = versionNumber();
				
				if (latestVersionString == nil)
					NSPrint([NSString stringWithFormat:@"...failed.\n\nError reading latest version number from HTTP header field.\n\n"]);
				else
				{
					if (versionNumberCompare(currentVersionString, latestVersionString) == NSOrderedAscending)
					{
						NSPrint([NSString stringWithFormat:@"...update found! (latest: %@  current: %@)\n\n", latestVersionString, currentVersionString]);
						NSPrint([NSString stringWithFormat:@"Navigate to the following URL to see the release notes and download the latest version:\n\n%@?currentversion=%@\n\n", kAppSiteURLPrefix, currentVersionString]);
						char inputChar;
						while(inputChar != 'y' && inputChar != 'n' && inputChar != 'Y' && inputChar != 'N' && inputChar != '\n')
						{
							NSPrint(@"Do you want to navigate to this URL now? [y/n] ");
							scanf("%s&*c",&inputChar);
						}
						
						if (inputChar == 'y' || inputChar == 'Y')
							[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:[NSString stringWithFormat: @"%@?currentversion=%@", kAppSiteURLPrefix, currentVersionString]]];
					}
					else
						NSPrint([NSString stringWithFormat:@"...you're up to date! (latest: %@  current: %@)\n\n", latestVersionString, currentVersionString]);
				}
			}
		}
		else
		{
			NSPrint([NSString stringWithFormat:@"...connection failed.\n\nError: - %@ %@\n\n",
				[error localizedDescription],
				[[error userInfo] objectForKey:NSErrorFailingURLStringKey]]);
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print all calendars
	// ------------------------------------------------------------------
	else if ([arg_output isEqualToString:@"calendars"])
	{
		NSArray *allCalendars = [[CalCalendarStore defaultCalendarStore] calendars];
		
		CalCalendar *cal;
		for (cal in allCalendars)
		{
			NSPrint([NSString stringWithFormat: @"* %@\n  uid: %@\n", [cal title], [cal uid]]);
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print all EVENTS today
	// ------------------------------------------------------------------
	else if ([arg_output isEqualToString:@"eventsToday"])
	{
		// get all calendars
		NSMutableArray *allCalendars = [NSMutableArray arrayWithCapacity:10];
		[allCalendars addObjectsFromArray: [[CalCalendarStore defaultCalendarStore] calendars]];
		
		// filter calendars to use:
		// - first include what has been specified to be included,
		// - then exclude what has been specified to be excluded
		if (arg_includeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(uid IN %@) OR (title IN %@)", arg_includeCals, arg_includeCals]];
		if (arg_excludeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(NOT(uid IN %@)) AND (NOT(title IN %@))", arg_excludeCals, arg_excludeCals]];
		
		// get start and end dates for predicate
		NSCalendarDate *now = [NSCalendarDate calendarDate];
		NSCalendarDate *todayStart = [NSCalendarDate dateWithYear:[now yearOfCommonEra] month:[now monthOfYear] day:[now dayOfMonth] hour:0 minute:0 second:0 timeZone:[now timeZone]];
		NSCalendarDate *todayEnd = [NSCalendarDate dateWithYear:[now yearOfCommonEra] month:[now monthOfYear] day:[now dayOfMonth] hour:23 minute:59 second:59 timeZone:[now timeZone]];
		
		// make predicate for getting all events for today + use it to get the events
		NSPredicate *eventsTodayPredicate = [CalCalendarStore eventPredicateWithStartDate:(arg_includeOnlyEventsFromNowOn)?now:todayStart endDate:todayEnd calendars:allCalendars];
		NSArray *eventsToday = [[CalCalendarStore defaultCalendarStore] eventsWithPredicate:eventsTodayPredicate];
		
		// print the events out
		if (arg_separateCals)
		{
			// separated by calendar
			BOOL titlePrintedForCurrentCalendar;
			BOOL currentIsFirstPrintedCalendar = YES;
			CalCalendar *cal;
			for (cal in allCalendars)
			{
				titlePrintedForCurrentCalendar = NO;
				CalEvent *event;
				for (event in eventsToday)
				{
					if ([[cal uid] isEqualToString: [[event calendar] uid]])
					{
						if (!titlePrintedForCurrentCalendar)
						{
							if (!currentIsFirstPrintedCalendar)
								NSPrint(@"\n");
							NSPrint([NSString stringWithFormat: @"%@:\n------------------------\n", [cal title]]);
							titlePrintedForCurrentCalendar = YES;
							currentIsFirstPrintedCalendar = NO;
						}
						printCalEvent(event, EVENTS_TODAY_NOCALENDAR,
										arg_timeFormatStr, arg_dateTimeFormatStr,
										arg_includedEventProperties, arg_excludedEventProperties);
					}
				}
			}
		}
		else
		{
			// not separated by calendar (calendar specified at the end of the title)
			CalEvent *event;
			for (event in eventsToday)
			{
				printCalEvent(event, EVENTS_TODAY,
								arg_timeFormatStr, arg_dateTimeFormatStr,
								arg_includedEventProperties, arg_excludedEventProperties);
			}
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	// print all uncompleted TASKS
	// ------------------------------------------------------------------
	else if ([arg_output isEqualToString:@"uncompletedTasks"])
	{
		// get all calendars
		NSMutableArray *allCalendars = [NSMutableArray arrayWithCapacity:10];
		[allCalendars addObjectsFromArray: [[CalCalendarStore defaultCalendarStore] calendars]];
		
		// filter calendars to use:
		// - first include what has been specified to be included,
		// - then exclude what has been specified to be excluded
		if (arg_includeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(uid IN %@) OR (title IN %@)", arg_includeCals, arg_includeCals]];
		if (arg_excludeCals != nil)
			[allCalendars filterUsingPredicate: [NSPredicate predicateWithFormat:@"(NOT(uid IN %@)) AND (NOT(title IN %@))", arg_excludeCals, arg_excludeCals]];
		
		// make predicate for uncompleted tasks in all calendars and use it to get the tasks
		NSPredicate *uncompletedTasksPredicate = [CalCalendarStore taskPredicateWithUncompletedTasks:allCalendars];
		NSArray *uncompletedTasks = [[CalCalendarStore defaultCalendarStore] tasksWithPredicate:uncompletedTasksPredicate];
		
		// sort the tasks by priority
		uncompletedTasks = [uncompletedTasks sortedArrayUsingFunction:prioritySort context:NULL];
		
		
		// print out the tasks
		if (arg_separateCals)
		{
			// separated by calendar
			BOOL titlePrintedForCurrentCalendar;
			BOOL currentIsFirstPrintedCalendar = YES;
			CalCalendar *cal;
			for (cal in allCalendars)
			{
				titlePrintedForCurrentCalendar = NO;
				CalTask *task;
				for (task in uncompletedTasks)
				{
					if ([[cal uid] isEqualToString: [[task calendar] uid]])
					{
						if (!titlePrintedForCurrentCalendar)
						{
							if (!currentIsFirstPrintedCalendar)
								NSPrint(@"\n");
							NSPrint([NSString stringWithFormat: @"%@:\n------------------------\n", [cal title]]);
							titlePrintedForCurrentCalendar = YES;
							currentIsFirstPrintedCalendar = NO;
						}
						printCalTask(task, TASKS_NOCALENDAR,
										arg_timeFormatStr, arg_dateTimeFormatStr,
										arg_includedTaskProperties, arg_excludedTaskProperties);
					}
				}
			}
		}
		else
		{
			// not separated by calendar (calendar specified at the end of the title)
			CalTask *task;
			for (task in uncompletedTasks)
			{
				printCalTask(task, TASKS,
								arg_timeFormatStr, arg_dateTimeFormatStr,
								arg_includedTaskProperties, arg_excludedTaskProperties);
			}
		}
	}
	// ------------------------------------------------------------------
	// ------------------------------------------------------------------
	else
	{
		NSPrint(@"\n");
		NSPrint([NSString stringWithFormat: @"USAGE: %s <options> <output_type>\n", argv[0]]);
		NSPrint(@"\n");
		NSPrint(@"<output_type> specifies what to output.\n");
		NSPrint(@"Possible values for it are:\n");
		NSPrint(@"  'eventsToday'      (outputs events occurring today)\n");
		NSPrint(@"  'uncompletedTasks' (outputs uncompleted tasks)\n");
		NSPrint(@"  'calendars'        (outputs all calendars)\n");
		NSPrint(@"\n");
		NSPrint(@"Options:\n");
		NSPrint(@"\n");
		NSPrint(@"  -V    (or: --version) Print version number and exit.\n");
		NSPrint(@"\n");
		NSPrint(@"  -u    (or: --checkForUpdates) Check for updates.\n");
		NSPrint(@"\n");
		NSPrint(@"  -s    Separate events/tasks by calendar.\n");
		NSPrint(@"\n");
		NSPrint(@"  -tf   (or: --timeFormat)\n");
		NSPrint(@"        Specify the formatting of times in the output.\n");
		NSPrint(@"        See http://tinyurl.com/5qvjrx [apple.com] for\n");
		NSPrint(@"        documentation on the syntax to use.\n");
		NSPrint(@"\n");
		NSPrint(@"  -dtf  (or: --dateTimeFormat)\n");
		NSPrint(@"        Specify the formatting of dates and times in the output.\n");
		NSPrint(@"        See http://tinyurl.com/5qvjrx [apple.com] for\n");
		NSPrint(@"        documentation on the syntax to use.\n");
		NSPrint(@"\n");
		NSPrint(@"  -n    (or: --includeOnlyEventsFromNowOn)\n");
		NSPrint(@"        If the output_type value 'eventsToday' is\n");
		NSPrint(@"        used, only output events that occur between the\n");
		NSPrint(@"        current date and the end of the day (as opposed to\n");
		NSPrint(@"        events that occur between the start and end of the\n");
		NSPrint(@"        day, as by default.)\n");
		NSPrint(@"\n");
		NSPrint(@"  -ic   (or: --includeCals) Which calendars to include.\n");
		NSPrint(@"        Must be a comma-separated list of calendar UIDs (which\n");
		NSPrint(@"        you can find out by using the 'calendars' output parameter.)\n");
		NSPrint(@"        You can also use titles, but remember that they may not be\n");
		NSPrint(@"        unique (as opposed to UIDs.)\n");
		NSPrint(@"        The -i and -e parameters will be handled in the order of\n");
		NSPrint(@"        first include, then exclude.\n");
		NSPrint(@"\n");
		NSPrint(@"  -ec   (or: --excludeCals) Which calendars to exclude.\n");
		NSPrint(@"        Must be a comma-separated list of calendar UIDs (which\n");
		NSPrint(@"        you can find out by using the 'calendars' output parameter.)\n");
		NSPrint(@"        You can also use titles, but remember that they may not be\n");
		NSPrint(@"        unique (as opposed to UIDs.)\n");
		NSPrint(@"        The -i and -e parameters will be handled in the order of\n");
		NSPrint(@"        first include, then exclude.\n");
		NSPrint(@"\n");
		NSPrint(@"  -iep  (or: --includeEventProps)\n");
		NSPrint(@"        Which event properties to include in the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. Possible\n");
		NSPrint(@"        property names are: location, url, notes, startDate\n");
		NSPrint(@"        and endDate.");
		NSPrint(@"\n");
		NSPrint(@"  -eep  (or: --excludeEventProps)\n");
		NSPrint(@"        Which event properties to exclude from the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. A value of \"*\"\n");
		NSPrint(@"        will exclude all properties and make sure just the title\n");
		NSPrint(@"        is printed.\n");
		NSPrint(@"\n");
		NSPrint(@"  -itp  (or: --includeTaskProps)\n");
		NSPrint(@"        Which task properties to include in the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. Possible\n");
		NSPrint(@"        property names are: url, notes, dueDate and priority.\n");
		NSPrint(@"\n");
		NSPrint(@"  -etp  (or: --excludeTaskProps)\n");
		NSPrint(@"        Which task properties to exclude from the output. Must be\n");
		NSPrint(@"        a comma-separated list property names. A value of \"*\"\n");
		NSPrint(@"        will exclude all properties and make sure just the title\n");
		NSPrint(@"        is printed.\n");
		NSPrint(@"\n");
		NSPrint([NSString stringWithFormat: @"Version %@\n", versionNumber()]);
		NSPrint(@"(c) 2008 Ali Rantakari, http://hasseg.org\n");
		NSPrint(@"\n");
	}
	
	
	[autoReleasePool release];
	return(0);
}